#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for searching, splitting, and inserting into a sorted list

Copied from https://github.com/python/cpython/blob/3.12/Lib/bisect.py with some modifications
noted in-line with `# EDIT`
"""


def insort_right(a, x, lo=0, hi=None, *, key=None):
    """Insert item x in list a, and keep it sorted assuming a is sorted.

    If x is already in a, insert it to the right of the rightmost x.

    :param a: The array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """
    if key is None:
        lo = bisect_right(a, x, lo, hi)
    else:
        lo = bisect_right(a, key(x), lo, hi, key=key)
    a.insert(lo, x)


def bisect_right(a, x, lo=0, hi=None, *, key=None):
    """Return the index where to insert item x in list a, assuming a is sorted.

    The return value i is such that all e in a[:i] have e <= x, and all e in
    a[i:] have e > x.  So if x already appears in the list, a.insert(i, x) will
    insert just after the rightmost x already there.

    :param a: The sorted array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """

    if lo < 0:
        # EDIT
        # raise ValueError('lo must be non-negative')
        lo = 0
    if hi is None:
        hi = len(a)
    # Note, the comparison uses "<" to match the
    # __lt__() logic in list.sort() and in heapq.
    if key is None:
        while lo < hi:
            mid = (lo + hi) // 2
            if x < a[mid]:
                hi = mid
            else:
                lo = mid + 1
    else:
        while lo < hi:
            mid = (lo + hi) // 2
            if x < key(a[mid]):
                hi = mid
            else:
                lo = mid + 1
    return lo


def insort_left(a, x, lo=0, hi=None, *, key=None):
    """Insert item x in list a, and keep it sorted assuming a is sorted.

    If x is already in a, insert it to the left of the leftmost x.

    :param a: The sorted array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """

    if key is None:
        lo = bisect_left(a, x, lo, hi)
    else:
        lo = bisect_left(a, key(x), lo, hi, key=key)
    a.insert(lo, x)


def bisect_left(a, x, lo=0, hi=None, *, key=None):
    """Return the index where to insert item x in list a, assuming a is sorted.

    The return value i is such that all e in a[:i] have e < x, and all e in
    a[i:] have e >= x.  So if x already appears in the list, a.insert(i, x) will
    insert just before the leftmost x already there.

    :param a: The sorted array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """

    if lo < 0:
        # EDIT
        # raise ValueError('lo must be non-negative')
        lo = 0
    if hi is None:
        hi = len(a)
    # Note, the comparison uses "<" to match the
    # __lt__() logic in list.sort() and in heapq.
    if key is None:
        while lo < hi:
            mid = (lo + hi) // 2
            if a[mid] < x:
                lo = mid + 1
            else:
                hi = mid
    else:
        while lo < hi:
            mid = (lo + hi) // 2
            if key(a[mid]) < x:
                lo = mid + 1
            else:
                hi = mid
    return lo


# Create aliases
bisect = bisect_right
insort = insort_right
